<?php

include_once __DIR__.'/constants.php';
include_once __DIR__.'/transactionHelper.php';


/**
 * formats the given string to be only 1 line
 * @param $output string the output
 * @return mixed|string
 */
function formatOutput($output) {

    global $responseNewLineString;

    $formattedOutput = $output;
    # backend requires output to be a string... so convert it if needed
    if (count($formattedOutput) === 0) {
        $formattedOutput = '';
    } else if (is_array($output)) {
        $formattedOutput = join($responseNewLineString, $formattedOutput);
        debug('output: ' . $formattedOutput);
    } else {
        //should be just 1 line string
        $formattedOutput = str_replace("\n", $responseNewLineString, $formattedOutput);
        debug('output: ' . $formattedOutput);
    }

    return $formattedOutput;
}

/**
 * handles the output from the test runner
 * @param $testResults array the test results (id, passed, hasCompiled, programExitCode, testResultCode, protocol)
 * @param $config object the config
 * @param $conn PDO the db connection
 * @param $rootDirName string the iteration root dir name
 * @param $transactionKey int the transaction key to end the transaction
 * @internal param string[] $output the output array from exec
 * @internal param int $return_var the return value from exec
 * @deprecated
 */
//TODO REMOVE
function handleExecOutput($testResults, $config, $conn, $rootDirName, $transactionKey)
{

    global $s_passed;
    global $s_hasCompiled;
    global $s_programExitCode;
    global $s_testResultCode;
    global $s_protocol;
    global $responseNewLineString;
    global $status_code_ArgumentInvalidOrMissing;
    global $status_code_InternalServerError;
    global $status_code_Ok;

    # release the transaction
    endTransaction($conn, $config, $transactionKey, $rootDirName);


    $response = [];

    //check if a test has some server error

    foreach ($testResults as $testResult) {

        $output = $testResult->data->content;
        $return_var = $testResult->return_var;

        if ($return_var === 101) {
            # unknown command argument
            # normally this can not happen because we checked the command argument
            $output = $testResult->data->content;

            if (is_array($output)) {
                $output = join($responseNewLineString, $output);
            }

            array_push($response, array(
                'responseCode' => $status_code_ArgumentInvalidOrMissing,
                'message' =>  "argument '" . 'test type' . "' was invalid or invalid number of arguments:" . $output,
                'data' => null
            ));

        } else if ($return_var === 100) {
            # unknown error occurred, more info is in the stdout (output from exec)
            if (is_array($output)) {
                $output = join($responseNewLineString, $output);
            }

            array_push($response, array(
                'responseCode' => $status_code_InternalServerError,
                'message' => 'unknown error: ' . $output,
                'data' => null
            ));
        } else { // all special return codes
            /*
             * 0 - all ok (conversation is in stdout)
             * 1 - output mismatch
             * 2 - some error (during program execution e.g. numberformat exception....)
             * 3 - timeout hit (conversation is in stdout)
             * 4 - memory limit hit (conversation is in stdout)
             * 5 - disk limit hit (conversation is in stdout)
             *
             * 50 - compile error (error is in stdout)
             * 60 - unknown / unsupported plang
             */

            # all are errors but the test was partly executed... and the exec out contains
            # the conversation protocol

            array_push($response, array(
                'responseCode' => $status_code_Ok,
                'message' => '',
                'data' => $testResult
            ));
        }

        outputResponse($response);
    }


}


/**
 * checks if the next dir is needed (we periodically change the root dir for the tests so that we can clear the old tests)
 * the interval specified in the config file - changeDirIntervalInS
 * ensure that the changeDirIntervalInS is less than the hardTimeoutInMs (else some tests might still be running during the deletion of the old dirs)
 * @param $hardGlobalTimeoutInMs int the global timeout in ms for every p langs
 * @return string the root dir name (a string of a timestamp)
 * @deprecated no longer used because we can now run multiple test with one request so the hard timeout is not for the whole request... this might be the reason we sometimes got a cannot initialize java vm error because dir not found or something...
 */
function checkNextDir($hardGlobalTimeoutInMs)
{

    global $s_timestampFile;
    global $s_lastRunDate;
    global $s_lastRunDir;
    global $timeZone;
    global $config;
//    global $s_changeDirIntervalInS;
    global $s_config_workingDirFullPath;
//    global $default_changeDirIntervalInS;
    global $status_code_FileSystemIssue;
    global $s_arg_hardTimeoutInMs ;
    global $status_code_InternalServerError;

    $jsonContent = file_get_contents($s_timestampFile);

    # pattern
    /*
     * {
     *      "lastRun": date,
     *      "lastDir": "dirName"
     * }
     */


    $lastRunObj = json_decode($jsonContent, TRUE, 512, JSON_FORCE_OBJECT);

//    $interval = intval($config[$s_changeDirIntervalInS]);

//    if ($interval === NULL || $interval <= 0) {
//        debug('change dir interval in s was null or could not be read, using default value: ' . $default_changeDirIntervalInS);
//        $interval = $default_changeDirIntervalInS;
//    }

    date_default_timezone_set($timeZone);
    $now = new DateTime('now');


    $lastRunDate = NULL;
    $lastRunDir = NULL;

    if ($lastRunObj === NULL) {
        debug('timestamp file could not be read... creating new one');

        $lastRunObj[$s_lastRunDate] = $lastRunDate = clone $now; //else we would get a reference
        $lastRunObj[$s_lastRunDir] = $lastRunDir = $now->getTimestamp();

        //create file
        if (file_put_contents($s_timestampFile, json_encode($lastRunObj, JSON_NUMERIC_CHECK | JSON_FORCE_OBJECT)) == FALSE) {
            output($status_code_FileSystemIssue, 'could not write to iteration timestamp file');
            return NULL;
            //exit(1); //if we exit then the cleanup won't be done!
        }

    } else {
        # we receive an array and no date time object so convert/create new
        $lastRunDate = new DateTime($lastRunObj[$s_lastRunDate]['date']);
        $lastRunDir = $lastRunObj[$s_lastRunDir];
    }

    $int = date_interval_create_from_date_string('+ ' . $interval . ' seconds');
    $lastRunDate->add($int);

    # if now is before last run date + interval = all ok
    if ($now < $lastRunDate) {
        # lastRunDate < now < lastRunDate + interval
        return $lastRunDir;

    } else if ($now > $lastRunDate) {
        # lastRunDate < lastRunDate + interval < now = next iteration (next dir)

        # if a test was started e.g. 1 second before the iteration was finished then we need to wait at least the hard timeout... better add 1 second

        date_add($lastRunDate, date_interval_create_from_date_string(sprintf('+%d seconds %d milliseconds', $interval, $hardGlobalTimeoutInMs)));

        if ($now < $lastRunDate) {
            # lastRunDate < lastRunDate + interval < now < lastRunDate + interval + $hardTimeout(+1s)
            # test could still be running
        } else {
            # lastRunDate < lastRunDate + interval < lastRunDate + interval + $hardTimeout(+1s) < now
            # now every test has finished... use next dir

            debug('finished iteration...');

            $rootDirName = date_timestamp_get($now);

            # write rootDirName and timestamp to the file
            $lastRunObj[$s_lastRunDate] = $now;
            $lastRunObj[$s_lastRunDir] = $rootDirName;

            if (file_put_contents($s_timestampFile, json_encode($lastRunObj, JSON_NUMERIC_CHECK | JSON_FORCE_OBJECT)) == FALSE) {
                output($status_code_FileSystemIssue, 'could not write to iteration timestamp file');
                exit(1);
            }

            # this is the root path for all tests but every iteration has its own root folder in that absolute root folder
            $config_workingDirFullPath = $config[$s_config_workingDirFullPath];

            #make the path end with udir separator
            if (!endsWith($config_workingDirFullPath, DIRECTORY_SEPARATOR)) {
                $config_workingDirFullPath = $config_workingDirFullPath . DIRECTORY_SEPARATOR;
            }

            if (!endsWith($lastRunDir, DIRECTORY_SEPARATOR)) {
                $lastRunDir = $lastRunDir . DIRECTORY_SEPARATOR;
            }

            $pathToDelete = $config_workingDirFullPath . $lastRunDir;

            debug('... deleting path ' . $pathToDelete);

            # we changed the dir ... delete old ones
            //TODO maybe remove all old directories but do not touch the new one
            deleteDir($pathToDelete);

            return $rootDirName;
        }
    }

    return $lastRunDir;
}


/**
 * creates (if not exists) the dirs used for the test
 * @param string $currentRootDirName the current root dir for this iteration
 * @param $tempUserDirName string the dir name to use/create
 * @return array|bool
 */
function ensureDirs($currentRootDirName, $tempUserDirName)
{

    global $s_config_workingDirFullPath;
    global $config;
    global $s_tempDirName;
    global $s_fullWorkingDirPath;
    global $timeZone;
    global $status_code_FileSystemIssue;
    global $isDebug;


    # this is the root path for all tests but every iteration has its own root folder in that absolute root folder
    $config_workingDirFullPath = $config[$s_config_workingDirFullPath];

    #make the path end with dir separator
    if (!endsWith($config_workingDirFullPath, DIRECTORY_SEPARATOR)) {
        $config_workingDirFullPath = $config_workingDirFullPath . DIRECTORY_SEPARATOR;
    }

    if (!endsWith($currentRootDirName, DIRECTORY_SEPARATOR)) {
        $currentRootDirName = $currentRootDirName . DIRECTORY_SEPARATOR;
    }


    #date_default_timezone_set($timeZone);
    #$workingDirTempName = date_timestamp_get(date_create());
    # timestamp is too bad... we use the db transaction id/key becasue we user microtime and a request distinction number
    # which should be unique for EVERY request
    $workingDirTempName = $tempUserDirName;


    $fullWorkingDirPath = $config_workingDirFullPath . $currentRootDirName . $workingDirTempName;

    $returnArray = array(
        $s_tempDirName => $workingDirTempName, //the name of the temp directory
        $s_fullWorkingDirPath => $fullWorkingDirPath, //the dir where the working files will be places
    );


    if (file_exists($fullWorkingDirPath) && is_dir($fullWorkingDirPath)) {
        //working dir already exists... ok
        return $returnArray;
    }

    # see https://stackoverflow.com/questions/3997641/why-cant-php-create-a-directory-with-777-permissions
    # why 0777 is not working...
    # not needed anymore
    # $old = umask(0);
    # create the temporarily dir for the files
    # TOOD maybe somehow make this better than 777 (because of firejail)
    # use 777 so that every user can access e.g. if we use firejail with starts as another user...
    if (mkdir($fullWorkingDirPath, 0755, TRUE) === FALSE) {

        if ($isDebug) {
            output($status_code_FileSystemIssue, 'could not create directory: ' . $fullWorkingDirPath);
        } else {
            output($status_code_FileSystemIssue, 'could not create directory');
        }

        #umask($old);
        return FALSE;
    }

    #umask($old);
    return $returnArray;
}

/**
 * creates all given files inside a temp dir
 * @param array $allFiles all files to create
 * Format: array[int] => {fileName: string, fileContent: string}
 * @global string $s_config_workingDirFullPath the config working dir path
 * @param string $fullWorkingDirPath the path (dir) where to put the files
 * @global string $config the config
 * @return bool FALSE: could not create files
 */
function createFiles($allFiles, $fullWorkingDirPath)
{

    global $s_fileName;
    global $s_fileContent;
    global $status_code_FileSystemIssue;
    global $isDebug;

    if ($isDebug) {
        debug("using working dir: " . $fullWorkingDirPath);
    }

    # now create all files for the test
    $count = 0;
    foreach ($allFiles as $file) {
        if (file_put_contents($fullWorkingDirPath . DIRECTORY_SEPARATOR . $file[$s_fileName], $file[$s_fileContent]) === FALSE) {
            output($status_code_FileSystemIssue, "could not create file nr. " . $count);
            return FALSE;
        }
        $count++;
    }

    return TRUE;
}

/**
 * @param array $allAssets all test assets
 * Format: array[int] => {fileName: string, fileContent: byte[]}
 * @param string $fullWorkingDirPath the path (dir) where to put the files
 * @return bool FALSE: could not create files
 */
function createTestAssets($allAssets, $fullWorkingDirPath)
{

    # do not output something here because we use this in the test case loop
    global $s_fileName;
    global $s_fileContent;
    global $status_code_FileSystemIssue;

    # now create all files for the test
    $count = 0;
    foreach ($allAssets as $file) {

        $content = base64_decode($file[$s_fileContent]);

        if ($content === FALSE) {
            //could not decode byte array...
            #output($status_code_FileSystemIssue, "could not decode create file nr. " . $count);
            return FALSE;
        }

        if (file_put_contents($fullWorkingDirPath . DIRECTORY_SEPARATOR . $file[$s_fileName], $content) === FALSE) {
            #output($status_code_FileSystemIssue, "could not create file nr. " . $count);
            return FALSE;
        }

        $count++;
    }
    return TRUE;
}

/**
 * removes the given dir with all content
 * taken from http://stackoverflow.com/questions/3349753/delete-directory-with-files-in-it
 * @param string $dirPath
 */
function deleteDir($dirPath)
{
    global $status_code_FileSystemIssue;

    if (!is_dir($dirPath)) {
        //this can happen if two different php processes try to delete the same dir??
        //this should not fail the test
        //output($status_code_FileSystemIssue, "could not delete temp dir");
        warn("could not delete temp dir (!is_dir(dir)): '" . $dirPath . "'");
        //don't exit() here because the transaction is not closed yet
        //return;
    }
    if (substr($dirPath, strlen($dirPath) - 1, 1) != '/') {
        $dirPath .= '/';
    }
    $files = glob($dirPath . '*', GLOB_MARK);
    foreach ($files as $file) {
        if (is_dir($file)) {
            deleteDir($file);
        } else {
            unlink($file);
        }
    }
    rmdir($dirPath);
}
